#replace_package_name_marker
__name__="dspy"
__package_name__="dspy-lite"
#replace_package_version_marker
__version__="3.1.3.post1"
__description__="DSPy with litellm replaced by litelm and numpy made optional"
__url__="https://github.com/kennethwolters/dspy-lite"
__author__="Kenneth Wolters"
__author_email__=""
